import math
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import date
from typing import Iterable, List

from tqdm import tqdm

from .repo import (get_all_pkg_names, get_pkg_downloads_info2,
                   get_pkg_github_info, get_pkg_pypi_info)
from .types import Package


def find_packages(name_search: str, show_progress_indicator) -> List[Package]:
    search_phrases = name_search.lower().split(',')
    def name_matches_phrases(pkg_name: str) -> bool:
        return all(
            search_phrase in pkg_name 
            for search_phrase in search_phrases
        )

    all_pkg_names = [name.lower() for name in get_all_pkg_names()]
    matching_pkg_names = [name for name in all_pkg_names if name_matches_phrases(name)]
    with_progress = progress_indicator(len(matching_pkg_names)) if show_progress_indicator else iter

    THREADS = 10
    with ThreadPoolExecutor(THREADS) as executor:
        futures = [
            executor.submit(get_package_info, pkg_name) 
            for pkg_name in matching_pkg_names
        ]

        return [
            future.result() 
            for future in with_progress(as_completed(futures))
        ]


def progress_indicator(total: int):
    def indicator_fn(iterable: Iterable) -> Iterable:
        return tqdm(iterable=iterable, total=total)
    return indicator_fn


def get_package_info(name: str) -> Package:
    pkg = Package(name=name)

    try:
        pypi_info = get_pkg_pypi_info(name) or {}
    except Exception as e:
        warnings.warn(f'failed to get pypi info for {name}: {e}', RuntimeWarning)
    else:
        info = pypi_info.get('info') or {}
        if summary := info.get('summary'):
            pkg.summary = summary
        pkg.version = info.get('version')

        vcs_urls = [
            (info.get('project_urls') or {}).get('Source'),
            info.get('home_page')
        ]
        vcs_url = next(iter(
            url
            for url in vcs_urls
            if url and any(
                vcs_name in url 
                for vcs_name in ['github', 'bitbucket', 'gitlab']
            )
        ), None)
        home_page = vcs_url or info.get('home_page')
        if home_page:
            pkg.home_page = home_page

        releases = pypi_info.get('releases') or {}
        if num_releases := len(releases):
            pkg.releases = num_releases
            if upload_times := [
                upload.get('upload_time') 
                for uploads in releases.values()
                for upload in uploads
            ]:
                pkg.last_release_date = max(upload_times)[:10]

    if pkg.releases:
        try:
            pkg.downloads = get_pkg_downloads_info2(name)
        except Exception as e:
            warnings.warn(f'failed to get downloads info for {name}: {e}', RuntimeWarning)

    if pkg.home_page and ('github.com' in pkg.home_page):
        try:
            github_info = get_pkg_github_info(pkg.home_page) or {}
        except Exception as e:
            warnings.warn(f'failed to get github info for {name}: {e}', RuntimeWarning)
        else:
            pkg.stars = github_info.get('stargazers_count')

    pkg.score = _get_score(pkg)

    return pkg


def _get_score(package: Package) -> int:
    if not package.last_release_date:
        return 0

    score: float = 0

    if package.downloads:
        score += math.log10(package.downloads)

    if package.stars:   
        score +=  math.log2(package.stars)

    days_from_last_release = (date.today() - date.fromisoformat(package.last_release_date)).days
    if days_from_last_release <= 30:
        score += 3
    elif days_from_last_release <= 60:
        score += 2
    elif days_from_last_release <= 120:
        score += 1
    elif days_from_last_release > 180:
        score -= 1    
    elif days_from_last_release > 360:
        score -= 2
    else:
        score -= 3    

    return round(score)



from dataclasses import dataclass
from typing import Optional, TypedDict, List, Dict

@dataclass
class Package:
    name: str
    summary: Optional[str] = None
    version: Optional[str] = None
    home_page: Optional[str] = None
    downloads: Optional[int] = None
    stars: Optional[int] = None
    releases: Optional[int] = None
    last_release_date: Optional[str] = None
    score: Optional[int] = None


class PypiProjectUrls(TypedDict):
    Source: Optional[str]

class PypiPackageInfo(TypedDict):
    summary: Optional[str]
    version: Optional[str] 
    project_urls: Optional[PypiProjectUrls]
    home_page: Optional[str]

class PypiReleaseUpload(TypedDict):
    upload_time: Optional[str]

class PypiEntry(TypedDict):
    info: Optional[PypiPackageInfo]
    releases: Optional[Dict[str, List[PypiReleaseUpload]]]


class GithubRepo(TypedDict):
    stargazers_count: Optional[int]


import json
import logging
from operator import attrgetter
from typing import List, Optional

import click
from tabulate import tabulate

from .github_auth import github_device_flow
from .repo import cache
from .service import find_packages
from .types import Package


@click.group()
def cli():
    pass


@cli.command()
def auth_github():
    """Log into GitHub"""
    with github_device_flow() as verif_codes:
        print(f'Please open {verif_codes["verification_uri"]} and enter code: {verif_codes["user_code"]}')
    print('Success')


def validate_pkg_name(ctx, param, value: str) -> str:
    MIN_LEN = 4
    if len(value) < MIN_LEN:
        raise click.BadParameter(f'name too short, min length={MIN_LEN}')

    return value   


@cli.command()
@click.argument('name-search', callback=validate_pkg_name)
@click.option('--limit', type=click.IntRange(min=1))
@click.option('--no-cache', is_flag=True, type=click.BOOL, default=False)
@click.option('--verbose', is_flag=True, type=click.BOOL, default=False)
@click.option('--json', "as_json", is_flag=True, type=click.BOOL, default=False)
def search(name_search: str, limit: int, no_cache: bool, verbose: bool, as_json: bool):
    """Search python package by name"""

    if no_cache:
        cache.clear()

    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)
        
    sorted_packages = list(reversed(sorted(find_packages(name_search, not as_json), key=attrgetter('score'))))
    if limit:
        sorted_packages = sorted_packages[:limit]

    print_func = _print_as_json if as_json else _print_as_text
    print_func(sorted_packages)


def _print_as_json(sorted_packages: List[Package]):
    for pkg in sorted_packages:
        print(json.dumps(pkg.__dict__))


def _print_as_text(sorted_packages: List[Package]):
    if pkg_count := len(sorted_packages):
        print(f'Found {pkg_count} packages:')
    else:
        print('No matching packages found')
        return 

    columns = ['name', 'downloads', 'summary', 'version', 'home_page', 'stars', 'releases', 'last_release_date']
    columns_max_width = {
        'name': 30,
        'summary': 50,
        'home_page': 50
    }

    def _enforse_max_width(val, max_width: Optional[int]):
        if max_width and type(val) == str and len(val) > max_width:
            return val[:max_width] + '...'
        else:
            return val


    print(tabulate([
        [
            _enforse_max_width(getattr(pkg, col), columns_max_width.get(col)) 
            for col in columns
        ]
        for pkg in sorted_packages
    ], headers=columns))


if __name__ == '__main__':
    cli()



import functools
import logging
from typing import Iterator, Optional

# import pypistats
import requests
from appdirs import user_cache_dir
from diskcache import Cache
from lxml import html

from .types import GithubRepo, PypiEntry
from .user_config import read_oauth_token

cache = Cache(user_cache_dir('pypi-client', 'PyPI'))


def get_all_pkg_names() -> Iterator[str]:   
    tree: html.HtmlElement = html.fromstring(_get_all_pkgs_html())
    return tree.xpath('//a/text()')


@cache.memoize()
def _get_all_pkgs_html() -> bytes:
    logging.debug('get_all_pkgs_html')
    response = requests.get("https://pypi.org/simple/")
    return response.content 


@cache.memoize()
def get_pkg_pypi_info(pkg_name: str) -> Optional[PypiEntry]:
    logging.debug(f'get_pkg_pypi_info for {pkg_name}')
    response = requests.get(f'https://pypi.org/pypi/{pkg_name}/json')
    if response.status_code == 404:
        return
    response.raise_for_status()
    return response.json()    


# @cache.memoize()
# def get_pkg_downloads_info1(pkg_name: str) -> Optional[dict]:
#     logging.debug(f'get_pkg_downloads_info1 for {pkg_name}')
#     try:
#         body = json.loads(pypistats.recent(pkg_name, format="json"))
#     except requests.exceptions.HTTPError as e:
#         if e.response.status_code == 404:
#             return
#         raise e
#     data = body.get('data') or {}
#     return data.get('last_month')


@cache.memoize()
def get_pkg_downloads_info2(pkg_name: str) -> Optional[int]:
    logging.debug(f'get_pkg_downloads_info2 for {pkg_name}')
    url = 'https://api.pepy.tech/api/v2/projects/' + pkg_name
    response = requests.get(url)
    if response.status_code == 404:
        return
    response.raise_for_status()
    body = response.json()
    return body.get('total_downloads')


@cache.memoize()
def get_pkg_github_info(pkg_repo: str) -> Optional[GithubRepo]:
    logging.debug(f'get_pkg_github_info for {pkg_repo}')
    repo_name = pkg_repo.strip('/').split('/')[-2:]
    url = 'https://api.github.com/repos/' + '/'.join(repo_name)
    response = requests.get(url, auth=('token', _get_github_oauth_token()))
    if response.status_code == 404:
        return
    response.raise_for_status()
    return response.json()


@functools.cache
def _get_github_oauth_token():
    return read_oauth_token()



import time
from contextlib import contextmanager
from typing import TypedDict, cast
from urllib import parse

import requests

from .user_config import write_oauth_token

CLIENT_ID = 'da5e9528b63f1bd10fd8'


class DeviceFlowVerificationCodes(TypedDict):
    device_code: str
    user_code: str
    verification_uri: str
    expires_in: int
    interval: int


@contextmanager
def github_device_flow():
    verif_codes = _get_verification_codes()
    yield verif_codes
    access_token = _wait_for_authorization(verif_codes['device_code'], verif_codes['expires_in'], verif_codes['interval'])
    write_oauth_token(access_token)


def _get_verification_codes() -> DeviceFlowVerificationCodes:
    res = requests.post('https://github.com/login/device/code', {'client_id': CLIENT_ID})
    res.raise_for_status()
    r = dict(parse.parse_qsl(res.text))
    return {
        'device_code': r['device_code'],
        'user_code': r['user_code'],
        'verification_uri': r['verification_uri'],
        'expires_in': int(r['expires_in']),
        'interval': int(r['interval']),
    }


def _wait_for_authorization(device_code: str, expires_in: int, sleep_interval: int) -> str:
    time.sleep(sleep_interval)
    expires_in -= sleep_interval

    while expires_in > 0:
        try:
            return _get_access_token(device_code)
        except Exception as e:
            time.sleep(sleep_interval)
            expires_in -= sleep_interval
    else:
        raise Exception('Verification code expired')


def _get_access_token(device_code: str) -> str:
    res = requests.post('https://github.com/login/oauth/access_token', {
        'client_id': CLIENT_ID,
        'device_code': device_code,
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code'
    })
    res.raise_for_status()
    r = dict(parse.parse_qsl(res.text))

    if 'error' in r:
        raise Exception(r['error_description'], r['error'])

    access_token = r.get('access_token')
    assert(access_token)
    return access_token



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['pypi_client']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'diskcache>=5.0.3,<6.0.0',
 'lxml>=4.6.1,<5.0.0',
 'pypistats>=0.12.1,<0.13.0',
 'requests>=2.24.0,<3.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'tqdm>=4.51.0,<5.0.0']

entry_points = \
{'console_scripts': ['pypi-client = pypi_client.cli:cli']}

setup_kwargs = {
    'name': 'pypi-client',
    'version': '0.1.2',
    'description': 'PyPI command-line tool',
    'long_description': 'PyPI client\n===========\n\n```\nUsage: pypi-client [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  auth-github  Log into GitHub\n  search       Search python package by name\n```',
    'author': 'Andrzej Bogdanowicz',
    'author_email': 'bahdanovich@rtbhouse.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abahdanovich/pypi-client',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
